/*******************************************************************************\
 * Copyright (c) 1997 by Markus Koelblin
 *
 * Permission is hereby granted, free of charge, to any person obtaining a copy
 * of this software and associated documentation files (the "Software"), to deal
 * in the Software without restriction, including without limitation the rights
 * to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
 * copies of the Software, and to permit persons to whom the Software is
 * furnished to do so, subject to the following conditions:
 * 
 * The above copyright notice and this permission notice shall be included in
 * all copies or substantial portions of the Software.
 * 
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
 * IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
 * FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT.  IN NO EVENT SHALL THE
 * AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
 * LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
 * OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
 * THE SOFTWARE.
\*******************************************************************************/

/*******************************************************************************\
 * listdlg.c                                                                   *
 *                                                                             *
 * Implements a list dialog window for rgb.txt color names and query results   *
 *                                                                             *
 * Author:  Markus Koelblin (Markus.Koelblin@hadiko.de) 1997                   *
\*******************************************************************************/

#include <X11/StringDefs.h>
#include <X11/Intrinsic.h>

#include <X11/Shell.h>
#include <X11/Xaw/Dialog.h>
#include <X11/Xaw/Form.h>
#include <X11/Xaw/List.h>
#include <X11/Xaw/Label.h>
#include <X11/Xaw/Command.h>
#include <X11/Xaw/Viewport.h>
#include <X11/Xaw/Scrollbar.h>

#include <stdio.h>
#include "rgbtxt.h"
#include "globals.h"



/* Offset for dialog position relative to the toplevel window */
#define DLG_OFFSET     24


/* Dialog data structure */
typedef struct tagColorListDialog
{
    Widget   caller, shell;
    Widget   viewport, list;
    Widget   close;

    RgbQuery *query;
    RgbBase  *rgbbase;
    int      index;
}
ColorListDialog;


static int popup_count = 0;


/*************************************************************************\
 *  get_toplevel()                                                       *
\*************************************************************************/

static Widget get_toplevel(Widget w)
{
    Widget p = w;

    while (w != None)
    {
	p = w;
	w = XtParent(w);
    }
    return p;
}


/*************************************************************************\
 *  FreeListbox()                                                        *
\*************************************************************************/

static void free_list_entries(Widget listbox)
{
    String  *strings;
    int     count, i;

    XtVaGetValues(listbox, XtNnumberStrings, &count, XtNlist, &strings, NULL);

    if (count > 1)
    {
	/* Free entry table of the list box */
	XtFree((XtPointer) strings);
    }
}

/*************************************************************************\
 *                                                                       *
\*************************************************************************/

static void set_list_entries(ColorListDialog* data)
{
    Widget  scrollbar;
    String  *list;
    int	    i, count;

    if (data->query == NULL)
    {
	/* Allocate memory for listbox rgb.txt entries */
	count = data->rgbbase->count; 
	list  = (String*) XtCalloc(sizeof(String*), count + 1);

	for (i = 0; i < count; i++)
	{
	    /* Setting the pointer for is enough here! */
	    list[i] = data->rgbbase->entries[i].name;
	}
    }
    else
    {
	/* Allocate memory for listbox query entries */
	count = data->query->count;
	list  = (String*) XtCalloc(sizeof(String*), count + 1);

	for (i = 0; i < count; i++)
	{
	    const int idx = data->query->indices[i];

	    /* Setting the pointer for is enough here! */
	    list[i] = data->rgbbase->entries[idx].name;
	}
    }

    /* Free old listbox entries (not really neccssary here) */
    free_list_entries(data->list);

    /* Assign new list to listbox widget */
    XawListChange(data->list, list, count, 0, True);

    if ((scrollbar = XtNameToWidget(data->viewport, "vertical")) != None)
    {
	/* Reset scrollbar postion */
	const float sb_top = 0.0;

	XawScrollbarSetThumb(scrollbar, sb_top, -1.0F);
	XtCallCallbacks(scrollbar, XtNjumpProc, (XtPointer) &sb_top);
    }
}


/*******************************************************************************\
 * Callbacks                                                                   *
\*******************************************************************************/

/*******************************************************************************\
 * close_CB()                                                                  *
\*******************************************************************************/

static void close_CB(Widget w, XtPointer client_data, XtPointer call_data)
{
    ColorListDialog *data = (ColorListDialog*) client_data;

    /* Hide dialogbox and clean up */
    XtPopdown(data->shell);

    free_list_entries(data->list);

    if (data->query)
    {
	/* Free indices and query structure itself */
	rgbtxt_free_query(data->query);
	free(data->query);
    }
    XtDestroyWidget(data->shell);
    XtSetSensitive(data->caller, True);
    XtFree((char*) data);
    popup_count--;
}

/*************************************************************************\
 * list_CB()                                                             *
 *                                                                       *
 * Listbox callback passes the selected string value to the              *
 * asciitext widget, ends the dialog on a mouse doubleclick              *
 * if the selection is valid by calling okButtonCB().                    *
\*************************************************************************/

static void list_CB(Widget w, XtPointer client_data, XtPointer call_data)
{
    XawListReturnStruct *item = (XawListReturnStruct*) call_data;
    ColorListDialog     *data = (ColorListDialog*)     client_data;

    if (item->list_index != XAW_LIST_NONE)
    {
	/* Store selection in dialog data structure */
	if (data->query == NULL)
	    data->index = item->list_index;
	else
	    data->index = data->query->indices[item->list_index];

	/* Show color in the toplevel window */
	show_listbox_selection(data->rgbbase, data->index);
    }
}


/*******************************************************************************\
 *  Dialog functions                                                           *
\*******************************************************************************/

/*******************************************************************************\
 *  create_listform()                                                          *
 *                                                                             *
 *  Build the listform widget tree add callbacks and accellerators             *
\*******************************************************************************/

static void create_listform(ColorListDialog* data, String caption)
{
    Widget form, listform, label;

    form = XtCreateManagedWidget("form", formWidgetClass, data->shell, NULL, 0);
    /*
     *  Put the listbox (Viewport plus List) into a seperate form
     */
    listform = XtVaCreateManagedWidget(
	"listform",     formWidgetClass, form,
	XtNright,       XtChainRight,
	XtNleft,        XtChainLeft,
	XtNtop,         XtChainTop,
	XtNwidth,       256,
	XtNheight,      256,
	XtNborderWidth, 0,
	NULL);

    label = XtVaCreateManagedWidget(
	"label",        labelWidgetClass, listform,
	XtNlabel,       caption,
	XtNright,       XtChainRight,
	XtNleft,        XtChainLeft,
	XtNbottom,      XtChainTop,
	XtNborderWidth, 0,
	NULL);

    data->viewport = XtVaCreateManagedWidget(
	"viewport",    viewportWidgetClass, listform,
	XtNfromVert,   label,
	XtNbottom,     XtChainBottom,
	XtNtop,        XtChainTop,
	XtNallowVert,  True,
	XtNallowHoriz, True,
	XtNuseBottom,  True,
	XtNuseRight,   True,
	XtNwidth,      128,
	XtNheight,     200,
	NULL);

    data->list= XtVaCreateManagedWidget(
	"list", listWidgetClass, data->viewport,
	XtNverticalList,   True,
	XtNforceColumns,   True,
	XtNdefaultColumns, 1,
	XtNnumberStrings,  0,
	XtNborderWidth,    1,
	NULL);
    /*
     * Create command buttons
     */
    data->close = XtVaCreateManagedWidget(
	"close", commandWidgetClass, form,
	XtNfromVert,  listform,
	XtNbottom,    XtChainBottom,
	NULL);

    /* Add callbacks for the listform and install the accelerators */
    XtAddCallback(data->list,  XtNcallback, list_CB,  (XtPointer) data);
    XtAddCallback(data->close, XtNcallback, close_CB, (XtPointer) data);
}


/*******************************************************************************\
 *  colorlist_dialog()                                                         *
 *                                                                             *
 *  If query is NULL the complete rgb.txt database is shown, otherwise a       *
 *  query popup is created.                                                    *
\*******************************************************************************/

void colorlist_dialog(Widget caller, Colormap colormap, RgbBase* base, RgbQuery* query)
{
    ColorListDialog*  popup;
    Position          x, y;
    Dimension         width;
    String            popup_title, list_title;
    Widget            parent = get_toplevel(caller);

    /* Try to allocate dialog data */
    if (!(popup = (ColorListDialog*) XtMalloc(sizeof(ColorListDialog))) )
	return;

    /* Show query or complete color database ? */
    if (query == NULL)
    {
	/* We want only one instance for the complete rgb.txt database ! */
	XtSetSensitive(caller, False);
	popup_title = XtNewString("colorlist");
	list_title  = XtNewString("Entries in rgb.txt:");
    }
    else
    {
	popup_title = XtNewString("querylist");
	list_title  = (String) malloc(strlen(query->pattern.name) + 32);
        sprintf(list_title, "Matches for %s:", query->pattern.name);
    }

    /* Setu popup data, create a dialog shell */
    popup->rgbbase = base;
    popup->query   = query;
    popup->caller  = caller;
    popup->shell   = XtVaCreatePopupShell(popup_title, transientShellWidgetClass, parent, NULL);

    /* Insert the listform widget tree - adds also the callbacks */
    create_listform(popup, list_title);

    /* Free temporary title strings */
    XtFree(popup_title);
    XtFree(list_title);

    /* Set dialog position relative to the toplevel parent window */
    XtVaGetValues(parent, XtNx, &x, XtNy, &y, XtNwidth, &width, NULL);
    x += DLG_OFFSET * (popup_count + 1);
    y += DLG_OFFSET * (popup_count);
    popup_count++;
    XtVaSetValues(popup->shell, XtNx, x + width, XtNy, y, NULL);

    /* Fill the listbox */
    set_list_entries(popup);

    /* Show the colorlist dialogbox */
    XtPopup(popup->shell, 0);
    XSetWindowColormap(XtDisplay(parent), XtWindow(popup->shell), colormap);
}

/*******************************************************************************/






