/*******************************************************************************\
 * Copyright (c) 1997 by Markus Koelblin
 *
 * Permission is hereby granted, free of charge, to any person obtaining a copy
 * of this software and associated documentation files (the "Software"), to deal
 * in the Software without restriction, including without limitation the rights
 * to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
 * copies of the Software, and to permit persons to whom the Software is
 * furnished to do so, subject to the following conditions:
 * 
 * The above copyright notice and this permission notice shall be included in
 * all copies or substantial portions of the Software.
 * 
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
 * IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
 * FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT.  IN NO EVENT SHALL THE
 * AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
 * LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
 * OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
 * THE SOFTWARE.
\*******************************************************************************/

/*******************************************************************************\
 * rgbtxt.c                                                                    *
 *                                                                             *
 * Implements X11 color database routines (rgb.txt)                            *
 *                                                                             *
 * Author:  Markus Koelblin (Markus.Koelblin@hadiko.de) 1997                   *
\*******************************************************************************/

#include <assert.h>
#include <ctype.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include "rgbtxt.h"


#ifndef NDEBUG
#define DEBUG_RGBTXT_PRINT_ENTRY(en) fprintf(stderr, "%d %d %d %s\n", \
					     en->red, en->green, en->blue, en->name);
#else
#define DEBUG_RGBTXT_PRINT_ENTRY(e)
#endif


char rgbtxt_errors[RGBTXT_ERROR_MAX][48] =
{
    "Can't open X color database ",
    "Can't allocate memory for database entries ",
    "Parse error while reading the X color database ",
};


/********************************************************************************\
 * rgbtxt_create_db()                                                           *
 *                                                                              *
 * Open and read X color database rgb.txt - returns RgbBase entries             *
\********************************************************************************/

int rgbtxt_create_db(RgbBase* base, char* filename)
{
    RgbTxtEntry  *entry;
    FILE         *file;
    char         line[RGBTXT_LINE_MAX];
    int          i, count, err_count;

    /* Try to open the color database */
    if ((file = fopen(filename, "r")) == NULL)
	rgbtxt_error(RGBTXT_OPEN_ERROR, filename);

    /* First count entries */
    count = 0;
    while (fgets(line, RGBTXT_LINE_MAX, file))
	count++;
    fclose(file);

    /* Re-open the color database */
    if ((file = fopen(filename, "r")) == NULL)
	rgbtxt_error(RGBTXT_OPEN_ERROR, filename);

    /* Allocate memory for entries */
    entry = (RgbTxtEntry*) malloc((count+1) * sizeof(RgbTxtEntry));

    if (entry == NULL)
	rgbtxt_error(RGBTXT_ALLOC_ERROR, NULL);

    /* Now read color entries */
    base->entries = entry;
    i = err_count = 0;

    while (fgets(line, RGBTXT_LINE_MAX, file))
    {
	char *cptr;
	int  r, g, b;
	int  nread;

	/* Only simple comment stripping */
	if (line[0] == '!')
	    continue;

	nread = sscanf(line, "%d %d %d", &r, &g, &b);

	entry->red   = (unsigned short) r;
	entry->green = (unsigned short) g;
	entry->blue  = (unsigned short) b;

	/* Get color name */
	cptr = line;
	while (!isalpha(*cptr) && *cptr != '\0')
	    cptr++;
	strcpy(entry->name, cptr);
	if ((cptr = strchr(entry->name, '\n')) != NULL)
	    *cptr = '\0';

	if (nread == 3)
	{
	    assert(i < count);
	    DEBUG_RGBTXT_PRINT_ENTRY(entry);
	    entry++;
	    i++;
	}
	else
	    err_count++;
    }
    base->count = i;
    fclose(file);

    return err_count == 0;
}


/********************************************************************************\
 * rgbtxt_free_db()                                                             *
 *                                                                              *
 * Free RgbBase entries                                                         *
\********************************************************************************/

void rgbtxt_free_db(RgbBase* base)
{
    assert(base->count);
    free(base->entries);
}


/********************************************************************************\
 * rgbtxt_match_name()                                                          *
 *                                                                              *
 * Convert rgb values to color name                                             *
\********************************************************************************/

int rgbtxt_match_name(RgbBase        *base,
		      char           **name,
		      unsigned short red,
		      unsigned short green,
		      unsigned short blue )
{
    RgbTxtEntry  *entry = base->entries;
    const int    count  = base->count;

    if (entry)
    {
	int i;

	*name =  entry->name;

	for (i = 0; i < count; i++)
	{
	    if (red == entry->red && green == entry->green && blue == entry->blue)
	    {
		*name = entry->name;
		return 1;
	    }
	    entry++;
	}
    }
    return 0;
}


/********************************************************************************\
 * rgbtxt_guess_name()                                                          *
 *                                                                              *
 * Tolerant conversion from rgb values to color name                            *
\********************************************************************************/

int rgbtxt_guess_name(RgbBase        *base,
		      char           **name,
		      unsigned short red,
		      unsigned short green,
		      unsigned short blue,
		      short          tolerance)
{
    RgbTxtEntry  *entry = base->entries;
    const int    count  = base->count;
    int          delta  = tolerance;
    int          i;

    if (entry)
    {
	*name = entry->name;

	for (i = 0; i < count; i++)
	{
	    int curr = abs(red - entry->red);

	    curr += abs(green - entry->green);
	    curr += abs(blue  - entry->blue);
 
	    if (curr < delta)
	    {
		*name = entry->name;
		delta = curr;
	    }
	    entry++;
	}
    }
    return delta < tolerance;
}


/********************************************************************************\
 * rgbtxt_create_query()                                                        *
 *                                                                              *
 * Create a list of matching colors for given RGB values and tolerance.         *
\********************************************************************************/

int rgbtxt_create_query(RgbBase *base, RgbQuery *query)
{
    RgbTxtEntry  *entry = base->entries;
    const int    count  = base->count;
    int          *list  = malloc(count * sizeof(int));
    int          nhits  = 0;

    if (entry && list)
    {
	const unsigned short red   = query->pattern.red;
	const unsigned short green = query->pattern.green;
	const unsigned short blue  = query->pattern.blue;
	const short          delta = query->tolerance;
	int i;

	for (i = 0; i < count; i++)
	{
	    int curr = abs(red - entry->red);

	    curr += abs(green - entry->green);
	    curr += abs(blue  - entry->blue);
 
	    if (curr < delta)
	    {
		/* Add index to matched item to query list */
		list[nhits] = i;
		nhits++;
	    }
	    entry++;
	}
    }

    if (nhits == 0)
	free(list);
    query->indices = list;
    query->count   = nhits;

    return nhits;
}

/********************************************************************************\
 * rgbtxt_free_query()                                                          *
 *                                                                              *
 * Free memory of query structure and index list.                               *
\********************************************************************************/

void rgbtxt_free_query(RgbQuery *query)
{
    assert(base->count);
    free(query->indices);
}


/********************************************************************************\
 * rgbtxt_error()                                                               *
 *                                                                              *
 * Error handler                                                                *
\********************************************************************************/

void rgbtxt_error(int err_nr, char* msg)
{
    assert(err_nr >= 0 && err_nr < RGBTXT_ERROR_MAX);

    fprintf(stderr, "%s ", rgbtxt_errors[err_nr]);
    if (msg) fputs(msg, stderr);
    fputs("\n", stderr);
    exit(err_nr + 1);
}

/*******************************************************************************/

